package meerkat.bulletinboard;

import com.google.common.util.concurrent.*;
import com.google.protobuf.ByteString;
import meerkat.comm.CommunicationException;
import meerkat.comm.MessageInputStream;
import meerkat.comm.MessageInputStream.MessageInputStreamFactory;
import meerkat.comm.MessageOutputStream;
import meerkat.crypto.concrete.SHA256Digest;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;
import meerkat.util.BulletinBoardUtils;

import javax.ws.rs.NotFoundException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 15-Mar-16.
 * This client is to be used mainly for testing.
 * It wraps a BulletinBoardServer in an asynchronous client.
 * This means the access to the server is direct (via method calls) instead of through a TCP connection.
 * The client implements both synchronous and asynchronous method calls, but calls to the server itself are performed synchronously.
 */
public class LocalBulletinBoardClient implements SubscriptionAsyncBulletinBoardClient{

    private final BulletinBoardServer server;
    private final ListeningScheduledExecutorService executorService;
    private final BatchDigest digest;
    private final int subsrciptionDelay;

    /**
     * Initializes an instance of the client
     * @param server an initialized Bulletin Board Server instance which will perform the actual processing of the requests
     * @param threadNum is the number of concurrent threads to allocate for the client
     * @param subscriptionDelay is the required delay between subscription calls in milliseconds
     */
    public LocalBulletinBoardClient(BulletinBoardServer server, int threadNum, int subscriptionDelay) {
        this.server = server;
        this.executorService = MoreExecutors.listeningDecorator(Executors.newScheduledThreadPool(threadNum));
        this.digest = new GenericBatchDigest(new SHA256Digest());
        this.subsrciptionDelay = subscriptionDelay;
    }

    private class MessagePoster implements Callable<Boolean> {

        private final BulletinBoardMessage msg;

        public MessagePoster(BulletinBoardMessage msg) {
            this.msg = msg;
        }


        @Override
        public Boolean call() throws Exception {
            return server.postMessage(msg).getValue();
        }

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg, FutureCallback<Boolean> callback) {

        Futures.addCallback(executorService.submit(new MessagePoster(msg)), callback);

        digest.update(msg.getMsg());
        return digest.digestAsMessageID();

    }

    private class CompleteBatchPoster implements Callable<Boolean> {

        private final CompleteBatch completeBatch;

        public CompleteBatchPoster(CompleteBatch completeBatch) {
            this.completeBatch = completeBatch;
        }


        @Override
        public Boolean call() throws Exception {

            if (!server.beginBatch(completeBatch.getBeginBatchMessage()).getValue())
                return false;

            int i=0;
            for (BatchData data : completeBatch.getBatchDataList()){

                BatchMessage message = BatchMessage.newBuilder()
                        .setSignerId(completeBatch.getSignature().getSignerId())
                        .setBatchId(completeBatch.getBeginBatchMessage().getBatchId())
                        .setSerialNum(i)
                        .setData(data)
                        .build();

                if (!server.postBatchMessage(message).getValue())
                    return false;

                i++;
            }

            return server.closeBatchMessage(completeBatch.getCloseBatchMessage()).getValue();
        }

    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {

        Futures.addCallback(executorService.schedule(new CompleteBatchPoster(completeBatch), subsrciptionDelay, TimeUnit.MILLISECONDS), callback);

        digest.update(completeBatch);
        return digest.digestAsMessageID();

    }

    private class BatchBeginner implements Callable<Boolean> {

        private final BeginBatchMessage msg;

        public BatchBeginner(BeginBatchMessage msg) {
            this.msg = msg;
        }


        @Override
        public Boolean call() throws Exception {
            return server.beginBatch(msg).getValue();
        }

    }

    @Override
    public void beginBatch(BeginBatchMessage beginBatchMessage, FutureCallback<Boolean> callback) {
        Futures.addCallback(executorService.submit(new BatchBeginner(beginBatchMessage)), callback);
    }

    private class BatchDataPoster implements Callable<Boolean> {

        private final ByteString signerId;
        private final int batchId;
        private final List<BatchData> batchDataList;
        private final int startPosition;

        public BatchDataPoster(ByteString signerId, int batchId, List<BatchData> batchDataList, int startPosition) {
            this.signerId = signerId;
            this.batchId = batchId;
            this.batchDataList = batchDataList;
            this.startPosition = startPosition;
        }


        @Override
        public Boolean call() throws Exception {

            BatchMessage.Builder msgBuilder = BatchMessage.newBuilder()
                    .setSignerId(signerId)
                    .setBatchId(batchId);

            int i = startPosition;
            for (BatchData data : batchDataList){

                msgBuilder.setSerialNum(i)
                        .setData(data);

                if (!server.postBatchMessage(msgBuilder.build()).getValue())
                    return false;

                i++;

            }

            return true;

        }

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, int startPosition, FutureCallback<Boolean> callback) {
        postBatchData(ByteString.copyFrom(signerId), batchId, batchDataList, startPosition, callback);
    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {
        postBatchData(signerId, batchId, batchDataList, 0, callback);
    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, int startPosition, FutureCallback<Boolean> callback) {
        Futures.addCallback(executorService.submit(new BatchDataPoster(signerId, batchId, batchDataList, startPosition)), callback);
    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {
        postBatchData(signerId, batchId, batchDataList, 0, callback);
    }

    private class BatchCloser implements Callable<Boolean> {

        private final CloseBatchMessage msg;

        public BatchCloser(CloseBatchMessage msg) {
            this.msg = msg;
        }


        @Override
        public Boolean call() throws Exception {
            return server.closeBatchMessage(msg).getValue();
        }

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, FutureCallback<Boolean> callback) {
        Futures.addCallback(executorService.submit(new BatchCloser(closeBatchMessage)), callback);
    }

    private class RedundancyGetter implements Callable<Float> {

        private final MessageID msgId;

        public RedundancyGetter(MessageID msgId) {
            this.msgId = msgId;
        }


        @Override
        public Float call() throws Exception {

            MessageFilterList filterList = MessageFilterList.newBuilder()
                    .addFilter(MessageFilter.newBuilder()
                            .setType(FilterType.MSG_ID)
                            .setId(msgId.getID())
                            .build())
                    .build();

            ByteArrayOutputStream byteOutputStream = new ByteArrayOutputStream();
            MessageOutputStream<BulletinBoardMessage> outputStream = new MessageOutputStream<>(byteOutputStream);
            server.readMessages(filterList,outputStream);

            MessageInputStream<BulletinBoardMessage> inputStream =
                    MessageInputStreamFactory.createMessageInputStream(
                            new ByteArrayInputStream(byteOutputStream.toByteArray()),
                            BulletinBoardMessage.class);

            if (inputStream.isAvailable())
                return 1.0f;
            else
                return 0.0f;

        }

    }

    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {
        Futures.addCallback(executorService.submit(new RedundancyGetter(id)), callback);
    }

    private class MessageReader implements Callable<List<BulletinBoardMessage>> {

        private final MessageFilterList filterList;

        public MessageReader(MessageFilterList filterList) {
            this.filterList = filterList;
        }


        @Override
        public List<BulletinBoardMessage> call() throws Exception {

            ByteArrayOutputStream byteOutputStream = new ByteArrayOutputStream();
            MessageOutputStream<BulletinBoardMessage> outputStream = new MessageOutputStream<>(byteOutputStream);
            server.readMessages(filterList, outputStream);

            MessageInputStream<BulletinBoardMessage> inputStream =
                    MessageInputStreamFactory.createMessageInputStream(
                            new ByteArrayInputStream(byteOutputStream.toByteArray()),
                            BulletinBoardMessage.class);

            return inputStream.asList();

        }

    }

    @Override
    public void readMessages(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
        Futures.addCallback(executorService.submit(new MessageReader(filterList)), callback);
    }

    class SubscriptionCallback implements FutureCallback<List<BulletinBoardMessage>> {

        private MessageFilterList filterList;
        private final FutureCallback<List<BulletinBoardMessage>> callback;

        public SubscriptionCallback(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
            this.filterList = filterList;
            this.callback = callback;
        }

        @Override
        public void onSuccess(List<BulletinBoardMessage> result) {

            // Report new messages to user
            callback.onSuccess(result);

            MessageFilterList.Builder filterBuilder = filterList.toBuilder();

            // If any new messages arrived: update the MIN_ENTRY condition
            if (result.size() > 0) {

                // Remove last filter from list (MIN_ENTRY one)
                filterBuilder.removeFilter(filterBuilder.getFilterCount() - 1);

                // Add updated MIN_ENTRY filter (entry number is successor of last received entry's number)
                filterBuilder.addFilter(MessageFilter.newBuilder()
                        .setType(FilterType.MIN_ENTRY)
                        .setEntry(result.get(result.size() - 1).getEntryNum() + 1)
                        .build());

            }

            filterList = filterBuilder.build();

            // Reschedule job
            Futures.addCallback(executorService.submit(new MessageReader(filterList)), this);

        }

        @Override
        public void onFailure(Throwable t) {

            // Notify caller about failure and terminate subscription
            callback.onFailure(t);

        }
    }

    @Override
    public void subscribe(MessageFilterList filterList, long startEntry, FutureCallback<List<BulletinBoardMessage>> callback) {

        MessageFilterList subscriptionFilterList =
                filterList.toBuilder()
                    .addFilter(MessageFilter.newBuilder()
                            .setType(FilterType.MIN_ENTRY)
                            .setEntry(startEntry)
                            .build())
                    .build();

        Futures.addCallback(executorService.submit(new MessageReader(subscriptionFilterList)), new SubscriptionCallback(subscriptionFilterList, callback));

    }

    @Override
    public void subscribe(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
        subscribe(filterList, 0, callback);
    }

    private class CompleteBatchReader implements Callable<CompleteBatch> {

        private final BatchSpecificationMessage batchSpecificationMessage;

        public CompleteBatchReader(BatchSpecificationMessage batchSpecificationMessage) {
            this.batchSpecificationMessage = batchSpecificationMessage;
        }


        @Override
        public CompleteBatch call() throws Exception {

            final String[] TAGS_TO_REMOVE = {BulletinBoardConstants.BATCH_TAG, BulletinBoardConstants.BATCH_ID_TAG_PREFIX};

            CompleteBatch completeBatch = new CompleteBatch(BeginBatchMessage.newBuilder()
                    .setSignerId(batchSpecificationMessage.getSignerId())
                    .setBatchId(batchSpecificationMessage.getBatchId())
                    .build());

            ByteArrayOutputStream byteOutputStream = new ByteArrayOutputStream();
            MessageOutputStream<BatchData> batchOutputStream = new MessageOutputStream<>(byteOutputStream);
            server.readBatch(batchSpecificationMessage,batchOutputStream);

            MessageInputStream<BatchData> batchInputStream =
                    MessageInputStreamFactory.createMessageInputStream(
                            new ByteArrayInputStream(byteOutputStream.toByteArray()),
                            BatchData.class);

            completeBatch.appendBatchData(batchInputStream.asList());

            MessageFilterList filterList = MessageFilterList.newBuilder()
                    .addFilter(MessageFilter.newBuilder()
                            .setType(FilterType.TAG)
                            .setTag(BulletinBoardConstants.BATCH_TAG)
                            .build())
                    .addFilter(MessageFilter.newBuilder()
                            .setType(FilterType.TAG)
                            .setTag(BulletinBoardConstants.BATCH_ID_TAG_PREFIX + completeBatch.getBeginBatchMessage().getBatchId())
                            .build())
                    .addFilter(MessageFilter.newBuilder()
                            .setType(FilterType.SIGNER_ID)
                            .setId(completeBatch.getBeginBatchMessage().getSignerId())
                            .build())
                    .build();

            byteOutputStream = new ByteArrayOutputStream();
            MessageOutputStream<BulletinBoardMessage> messageOutputStream = new MessageOutputStream<>(byteOutputStream);
            server.readMessages(filterList,messageOutputStream);

            MessageInputStream<BulletinBoardMessage> messageInputStream =
                    MessageInputStreamFactory.createMessageInputStream(
                            new ByteArrayInputStream(byteOutputStream.toByteArray()),
                            BulletinBoardMessage.class);

            if (!messageInputStream.isAvailable())
                throw new NotFoundException("Batch does not exist");

            BulletinBoardMessage message = messageInputStream.readMessage();

            completeBatch.setBeginBatchMessage(BeginBatchMessage.newBuilder()
                    .addAllTag(BulletinBoardUtils.removePrefixTags(message, Arrays.asList(TAGS_TO_REMOVE)))
                    .setSignerId(message.getSig(0).getSignerId())
                    .setBatchId(Integer.parseInt(BulletinBoardUtils.findTagWithPrefix(message, BulletinBoardConstants.BATCH_ID_TAG_PREFIX)))
                    .build());

            completeBatch.setSignature(message.getSig(0));
            completeBatch.setTimestamp(message.getMsg().getTimestamp());

            return completeBatch;

        }

    }

    @Override
    public void readBatch(BatchSpecificationMessage batchSpecificationMessage, FutureCallback<CompleteBatch> callback) {
        Futures.addCallback(executorService.submit(new CompleteBatchReader(batchSpecificationMessage)), callback);
    }

    private class SyncQueryHandler implements Callable<SyncQueryResponse> {

        private final SyncQuery syncQuery;

        public SyncQueryHandler(SyncQuery syncQuery) {
            this.syncQuery = syncQuery;
        }


        @Override
        public SyncQueryResponse call() throws Exception {
            return server.querySync(syncQuery);
        }

    }

    @Override
    public void querySync(SyncQuery syncQuery, FutureCallback<SyncQueryResponse> callback) {
        Futures.addCallback(executorService.submit(new SyncQueryHandler(syncQuery)), callback);
    }

    /**
     * This method is a stub, since the implementation only considers one server, and that is given in the constructor
     * @param ignored is ignored
     */
    @Override
    public void init(BulletinBoardClientParams ignored) {}

    @Override
    public MessageID postMessage(BulletinBoardMessage msg) throws CommunicationException {

        try {

            MessagePoster poster = new MessagePoster(msg);
            poster.call();

            digest.update(msg);
            return digest.digestAsMessageID();

        } catch (Exception e) {
            return null;
        }

    }

    @Override
    public float getRedundancy(MessageID id) {

        try {

            RedundancyGetter getter = new RedundancyGetter(id);
            return getter.call();

        } catch (Exception e) {
            return -1.0f;
        }

    }

    @Override
    public List<BulletinBoardMessage> readMessages(MessageFilterList filterList) {

        try {

            MessageReader reader = new MessageReader(filterList);
            return reader.call();

        } catch (Exception e){
            return null;
        }

    }

    @Override
    public SyncQuery generateSyncQuery(GenerateSyncQueryParams GenerateSyncQueryParams) throws CommunicationException {
        return server.generateSyncQuery(GenerateSyncQueryParams);
    }

    @Override
    public void close() {
        try {
            server.close();
        } catch (CommunicationException ignored) {}
    }

}
